﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Text;
using System.Threading.Tasks;
using System.IO;
using Sandbox.Common;
using Sandbox.Common.Components;
using Sandbox.Common.ObjectBuilders;
using Sandbox.Definitions;
using Sandbox.ModAPI;
using Sandbox.ModAPI.Interfaces;
using Sandbox.Game.Entities;
using Sandbox.Game.Gui;
using Sandbox.Game;
using VRage.Common.Utils;
using VRageMath;
using VRage;
using VRage.ObjectBuilders;
using VRage.Components;
using VRage.ModAPI;
using VRage.Utils;
using VRage.Library.Utils;

using VRage.Game.ModAPI;
using VRage.Game.ModAPI.Interfaces;
using VRage.Game;

using Ingame = Sandbox.ModAPI.Ingame;


namespace Comfyfutons.AtmosphericDrag
{
    public class CompressionHeatedBlock
    {
        public IMySlimBlock block;


        private MyCubeGrid grid;
        private Vector3 primaryColor;
        float heat;
        public HashSet<int> heatShieldBlocks = new HashSet<int>();

        private Vector3 TARGET_COLOR = new Vector3(255, 0, 0);

        public CompressionHeatedBlock(IMySlimBlock block, MyCubeGrid grid, HashSet<int> heatShieldBlocks)
        {
            this.heatShieldBlocks = heatShieldBlocks;
            this.block = block;
            this.grid = grid;
            primaryColor = HSVtoRGB(block.GetColorMask());
            heat = 0;
        }

        public void heatBlock(float heatValue)
        {
            heat += heatValue;

            if (heat > 10000f)
            {
                float damage = heat - 10000f;
                damage = (damage * grid.GridSize * grid.GridSize) / 5f;
                heat = 10000f;
                if (heatShieldBlocks.Contains((int)block.MaxIntegrity))
                {
                    damageBlock(damage / 100);
                }
                else
                {
                    damageBlock(damage);
                }
            }
            //MyAPIGateway.Utilities.ShowNotification("Heating, Current Heat: " + heat);
        }

        public void coolBlock(float coolValue)
        {
            heat -= coolValue;
            if (heat < 0f)
                heat = 0f;
            if (heat <= 100f)
                grid.ColorBlocks(block.Position, block.Position, RGBtoHSV(primaryColor), false);
            else
            {
                double heatPercent = (Math.Asin((heat / 5000) - 1) / Math.PI) + 0.5f;
                Vector3 heatedColor = Vector3.Lerp(primaryColor, TARGET_COLOR, (float)heatPercent);
                //Vector3 heatedColor = Vector3.Lerp(primaryColor, TARGET_COLOR, heat / 10000f);
                grid.ColorBlocks(block.Position, block.Position, RGBtoHSV(heatedColor), true);
            }
            //MyAPIGateway.Utilities.ShowNotification("Cooling, Current Heat: " + heat);
        }

        public float getHeat()
        {
            return heat;
        }

        private void damageBlock(float damage)
        {
            IMyDestroyableObject damagedBlock = block as IMyDestroyableObject;
            IgnoreException(() => damagedBlock.DoDamage(damage, Sandbox.Common.ObjectBuilders.Definitions.MyDamageType.Environment, false));
        }

        // HSVtoRGB Method: Converts SE HSV values to 0-255 RGB values;
        public static Vector3 HSVtoRGB(Vector3 hsv)
        {
            //input
            float h = hsv.X * 360;
            float s = (hsv.Y + 1f) / 2f;
            float v = (hsv.Z + 1f) / 2f;


            //Usage Variables
            float c;
            float x;
            float m;
            Vector3 rgbPrime;
            Vector3 rgb;


            c = v * s;
            x = c * (1 - Math.Abs((h / 60) % 2 - 1));
            m = v - c;

            if (0 <= h && h < 60)
            {
                rgbPrime = new Vector3(c, x, 0);
            }
            else if (60 <= h && h < 120)
            {
                rgbPrime = new Vector3(x, c, 0);
            }
            else if (120 <= h && h < 180)
            {
                rgbPrime = new Vector3(0, c, x);
            }
            else if (180 <= h && h < 240)
            {
                rgbPrime = new Vector3(0, x, c);
            }
            else if (240 <= h && h < 300)
            {
                rgbPrime = new Vector3(x, 0, c);
            }
            else if (300 <= h && h < 360)
            {
                rgbPrime = new Vector3(c, 0, x);
            }
            else
                rgbPrime = new Vector3();

            rgb = new Vector3((rgbPrime.X + m) * 255, (rgbPrime.Y + m) * 255, (rgbPrime.Z + m) * 255);
            return rgb;
        }

        // RGBtoHSV Method: Converts 0-255 RGB values to SE HSV values;
        public static Vector3 RGBtoHSV(Vector3 rgb)
        {
            //input
            float r = rgb.X / 255;
            float g = rgb.Y / 255;
            float b = rgb.Z / 255;

            //Usage Variables
            float cMin;
            float cMax;
            float delt;
            Vector3 hsv = new Vector3();

            cMax = Math.Max(r, g);
            cMax = Math.Max(cMax, b);

            cMin = Math.Min(r, g);
            cMin = Math.Min(cMin, b);

            delt = cMax - cMin;


            // Hue Calculation
            if (delt == 0)
            {
                hsv.X = 0;
            }
            else if (cMax == r)
            {
                hsv.X = 60 * (((g - b) / delt) % 6);
            }
            else if (cMax == g)
            {
                hsv.X = 60 * (((b - r) / delt) + 2);
            }
            else if (cMax == b)
            {
                hsv.X = 60 * (((r - g) / delt) + 4);
            }
            else
                hsv.X = 0;

            // Saturation Calculation
            if (cMax == 0)
                hsv.Y = 0;
            else
                hsv.Y = delt / cMax;

            // Value Calculation
            hsv.Z = cMax;

            hsv.X = (float)Math.Round(hsv.X);

            hsv.X = hsv.X / 360;
            hsv.Y = (hsv.Y * 2f) - 1f;
            hsv.Z = (hsv.Z * 2f) - 1f;

            hsv.Y = (float)Math.Round(hsv.Y, 2);
            hsv.Z = (float)Math.Round(hsv.Z, 2);

            return hsv;
        }

        private void IgnoreException(Action act)
        {
            try
            {
                act.Invoke();
            }
            catch
            {
            }
        }

    }
}