﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Text;
using System.Threading.Tasks;
using System.IO;
using Sandbox.Common;
using Sandbox.Common.Components;
using Sandbox.Common.ObjectBuilders;
using Sandbox.Definitions;
using Sandbox.ModAPI;
using Sandbox.ModAPI.Interfaces;
using Sandbox.Game.Entities;
using Sandbox.Game.Gui;
using Sandbox.Game;
using VRage.Common.Utils;
using VRageMath;
using VRage;
using VRage.ObjectBuilders;
using VRage.Components;
using VRage.ModAPI;
using VRage.Utils;
using VRage.Library.Utils;

using Ingame = Sandbox.ModAPI.Ingame;
using VRage.Game;

namespace Comfyfutons.AtmosphericDrag
{
    [MyEntityComponentDescriptor(typeof(MyObjectBuilder_CubeGrid))]
    public class Drag : MyGameLogicComponent
    {
        // Member Variables
        private MyObjectBuilder_EntityBase objectBuilder;
        private MyCubeGrid grid;

        /* COMMENTED OUT TILL I HAVE TIME TO FIX... API CHANGED
        MyParticleEffect effect;
		*/
		
        private int slowUpdate = 0;
        private float atmosphere = 0;
        private float reEntryAtmosphere = 0;
        private int atmospheres = 0;

        private int countForwardBackward = 0;
        private int countLeftRight = 0;
        private int countUpDown = 0;

        private int blockCount = 0;

        private int skipCounter1 = 0;
        private int skipCounter10 = 0;
        private int skipCounter100 = 0;

        private const int SKIP_TICKS = 3;

        public const float MIN_ATMOSPHERE = 0.0f;
        public const float MAX_ATMOSPHERE = 0.82f;
        public const int COMPRESSION_DAMAGE_THRESHOLD = 250;

        private List<IMySlimBlock> forwardSurfaceBlocks = new List<IMySlimBlock>();
        private List<IMySlimBlock> backwardSurfaceBlocks = new List<IMySlimBlock>();
        private List<IMySlimBlock> upSurfaceBlocks = new List<IMySlimBlock>();
        private List<IMySlimBlock> downSurfaceBlocks = new List<IMySlimBlock>();
        private List<IMySlimBlock> leftSurfaceBlocks = new List<IMySlimBlock>();
        private List<IMySlimBlock> rightSurfaceBlocks = new List<IMySlimBlock>();

        private Dictionary<Vector3I, CompressionHeatedBlock> compressionHeatedBlocks = new Dictionary<Vector3I, CompressionHeatedBlock>();
        private HashSet<Vector3I> heatedBlockRemove = new HashSet<Vector3I>();
        public HashSet<int> heatShieldBlocks = new HashSet<int>();

        bool moving = false;

        // Initialize objects
        public override void Init(MyObjectBuilder_EntityBase objectBuilder)
        {
            this.objectBuilder = objectBuilder;
            grid = Entity as MyCubeGrid;
            Entity.NeedsUpdate |= MyEntityUpdateEnum.EACH_FRAME;
            addHeatShieldBlocks();
        }

        public override void UpdateAfterSimulation()
        {
            if (!isShipValid())
                return;

            /* Other update method using blockCount != blockCount
            if (blockCount != grid.BlocksCount)
            {
                updateSurfaceArea();
                blockCount = grid.BlocksCount;

            }
            */

            // Updates surface area when ship starts to move
            if (grid.Physics.LinearVelocity.Length() > 1 && !moving)
            {
                moving = true;
                updateSurfaceArea();
            }
            if (grid.Physics.LinearVelocity.Length() <= 1 && moving)
            {
                moving = false;
            }

			/* COMMENTED OUT TILL I HAVE TIME TO FIX... API CHANGED
            if (skipCounter1 == 0)
            {
                if(effect != null)
                {
                    if (effect.ID == 101 && grid.Physics.LinearVelocity.Length() < (COMPRESSION_DAMAGE_THRESHOLD - 50))
                    {
                        effect.Stop();
                        effect = null;
                    }
                    else
                    {
                        MatrixD locationMatrix = grid.WorldMatrix;
                        locationMatrix.Translation = grid.Physics.CenterOfMassWorld;
						effect.WorldMatrix = locationMatrix;
                    }
                }
            }
			*/


            if (skipCounter1 == 1 && grid.Physics.LinearVelocity.Length() > 1)
                calculateAndApplyDrag();

            skipCounter1++;
            if (skipCounter1 > 1)
            {
                skipCounter1 = 0;
            }
        }

        public override void UpdateAfterSimulation10()
        {

            if (!isShipValid())
                return;

            if (compressionHeatedBlocks.Count > 0 && skipCounter10 == 3)
            {
                foreach (var block in compressionHeatedBlocks)
                {
                    if (block.Value.block == null || block.Value.block.IsDestroyed)
                    {
                        heatedBlockRemove.Add(block.Key);
                        continue;
                    }
                    //CompressionHeatedBlock heatedBlock = block.Value;
                    //heatedBlock.coolBlock(100f);
                    block.Value.coolBlock(100f);
                }
            }

            if (grid.Physics.LinearVelocity.Length() > (float)COMPRESSION_DAMAGE_THRESHOLD && skipCounter10 == 1 && reEntryAtmosphere > 0)
            {
                calculateAndApplyCompressionHeat();
            }

            skipCounter10++;

            if (skipCounter10 >= 4)
                skipCounter10 = 0;
        }

        public override void UpdateAfterSimulation100()
        {
            if (skipCounter100 == 3)
            {
                foreach (var block in compressionHeatedBlocks)
                {
                    if (block.Value.block == null || block.Value.block.IsDestroyed)
                    {
                        heatedBlockRemove.Add(block.Key);
                        continue;
                    }
                    CompressionHeatedBlock heatedBlock = block.Value;
                    if (heatedBlock.getHeat() <= 0)
                    {
                        heatedBlockRemove.Add(block.Key);
                    }
                }
            }
            if (heatedBlockRemove.Count > 0 && skipCounter100 == 4)
            {
                foreach (var block in heatedBlockRemove)
                {
                    compressionHeatedBlocks.Remove(block);
                }
                heatedBlockRemove.Clear();
            }

            skipCounter100++;
            if (skipCounter100 > 4)
            {
                skipCounter100 = 0;
            }
        }

        public override MyObjectBuilder_EntityBase GetObjectBuilder(bool copy = false)
        {
            return copy ? (MyObjectBuilder_EntityBase)objectBuilder.Clone() : objectBuilder;
        }

        // updateSurfaceArea Method: Updates surface area
        private void updateSurfaceArea()
        {

            try
            {

                forwardSurfaceBlocks.Clear();
                backwardSurfaceBlocks.Clear();
                upSurfaceBlocks.Clear();
                downSurfaceBlocks.Clear();
                leftSurfaceBlocks.Clear();
                rightSurfaceBlocks.Clear();

                Vector3D Max = grid.Max;
                Vector3D Min = grid.Min;

                countForwardBackward = 0;
                countLeftRight = 0;
                countUpDown = 0;

                float gridSize = grid.GridSize;
                bool blockHit = false;

                //Calculate Forward Backward Surface Area
                for (int i = (int)Min.X; i <= Max.X; i++)
                {
                    for (int j = (int)Min.Y; j <= Max.Y; j++)
                    {
                        blockHit = false;
                        for (int k = (int)Min.Z; k <= Max.Z && !blockHit; k++)
                        {
                            IMySlimBlock tempBlock = grid.GetCubeBlock(new Vector3I(i, j, k));
                            if (tempBlock != null)
                            {
                                blockHit = true;
                                countForwardBackward++;
                                forwardSurfaceBlocks.Add(tempBlock);
                                break;
                            }
                        }
                        if (blockHit)
                        {
                            for (int k = (int)Max.Z; k >= Min.Z; k--)
                            {
                                IMySlimBlock tempBlock = grid.GetCubeBlock(new Vector3I(i, j, k));
                                if (tempBlock != null)
                                {
                                    backwardSurfaceBlocks.Add(tempBlock);
                                    break;
                                }
                            }
                        }
                    }
                }

                //Calculate Up Down Surface Area
                for (int i = (int)Min.X; i <= Max.X; i++)
                {
                    for (int j = (int)Min.Z; j <= Max.Z; j++)
                    {
                        blockHit = false;
                        for (int k = (int)Min.Y; k <= Max.Y && !blockHit; k++)
                        {
                            IMySlimBlock tempBlock = grid.GetCubeBlock(new Vector3I(i, k, j));
                            if (tempBlock != null)
                            {
                                blockHit = true;
                                countUpDown++;
                                upSurfaceBlocks.Add(tempBlock);
                                break;
                            }
                        }
                        if (blockHit)
                        {
                            for (int k = (int)Max.Y; k >= Min.Y; k--)
                            {
                                IMySlimBlock tempBlock = grid.GetCubeBlock(new Vector3I(i, k, j));
                                if (tempBlock != null)
                                {
                                    downSurfaceBlocks.Add(tempBlock);
                                    break;
                                }
                            }
                        }
                    }
                }

                //Calculate Left Right Surface Area
                for (int i = (int)Min.Y; i <= Max.Y; i++)
                {
                    for (int j = (int)Min.Z; j <= Max.Z; j++)
                    {
                        blockHit = false;
                        for (int k = (int)Min.X; k <= Max.X && !blockHit; k++)
                        {
                            IMySlimBlock tempBlock = grid.GetCubeBlock(new Vector3I(k, i, j));
                            if (tempBlock != null)
                            {
                                blockHit = true;
                                countLeftRight++;
                                leftSurfaceBlocks.Add(tempBlock);
                                break;
                            }
                        }
                        if (blockHit)
                        {
                            for (int k = (int)Max.X; k >= Min.X; k--)
                            {
                                IMySlimBlock tempBlock = grid.GetCubeBlock(new Vector3I(k, i, j));
                                if (tempBlock != null)
                                {
                                    rightSurfaceBlocks.Add(tempBlock);
                                    break;
                                }
                            }
                        }
                    }
                }

                /* V1.2 Surface Area Calculations
                    
                    
                Vector3D Max = grid.Max;
                Vector3D Min = grid.Min;
                Vector3D gridCenter = grid.Physics.CenterOfMassWorld;

                int box = Math.Max((int)Max.X, (int)Max.Y);
                box = Math.Max(box, (int)Max.Z);
                box = Math.Max(box, (int)Math.Abs(Min.X));
                box = Math.Max(box, (int)Math.Abs(Min.Y));
                box = Math.Max(box, (int)Math.Abs(Min.Z));

                countForwardBackward = 0;
                countLeftRight = 0;
                countUpDown = 0;

                float gridSize = grid.GridSize; 

                //Calculate Forward Backward Surface Area
                for (int i = -box; i <= box; i++)
                {
                    for (int j = -box; j <= box; j++)
                    {
                        Vector3D front = gridCenter + grid.WorldMatrix.Forward * gridSize * box;
                        Vector3D back = gridCenter + grid.WorldMatrix.Backward * gridSize * box;
                        if (i != 0)
                        {
                            front = front + grid.WorldMatrix.Left * gridSize * i;
                            back = back + grid.WorldMatrix.Left * gridSize * i;
                        }
                        if (j != 0)
                        {
                            front = front + grid.WorldMatrix.Up * gridSize * j;
                            back = back + grid.WorldMatrix.Up * gridSize * j;
                        }
                        if (grid.RayCastBlocks(front, back) != null)
                        {
                            countForwardBackward++;
                        }
                    }
                }

                // Calculate Left Right Surface Area
                for (int i = -box; i <= box; i++)
                {
                    for (int j = -box; j <= box; j++)
                    {
                        Vector3D left = gridCenter + grid.WorldMatrix.Left * gridSize * box;
                        Vector3D right = gridCenter + grid.WorldMatrix.Right * gridSize * box;
                        if (i != 0)
                        {
                            left = left + grid.WorldMatrix.Backward * gridSize * i;
                            right = right + grid.WorldMatrix.Backward * gridSize * i;
                        }
                        if (j != 0)
                        {
                            left = left + grid.WorldMatrix.Up * gridSize * j;
                            right = right + grid.WorldMatrix.Up * gridSize * j;
                        }
                        if (grid.RayCastBlocks(left, right) != null)
                        {
                            countLeftRight++;
                        }
                    }
                }

                // Calculate Up Down Surface Area
                for (int i = -box; i <= box; i++)
                {
                    for (int j = -box; j <= box; j++)
                    {
                        Vector3D up = gridCenter + grid.WorldMatrix.Up * gridSize * box;
                        Vector3D down = gridCenter + grid.WorldMatrix.Down * gridSize * box;
                        if (i != 0)
                        {
                            up = up + grid.WorldMatrix.Left * gridSize * i;
                            down = down + grid.WorldMatrix.Left * gridSize * i;
                        }
                        if (j != 0)
                        {
                            up = up + grid.WorldMatrix.Backward * gridSize * j;
                            down = down + grid.WorldMatrix.Backward * gridSize * j;
                        }
                        if (grid.RayCastBlocks(up, down) != null)
                        {
                            countUpDown++;
                        }

                    }
                }
                */
            }
            catch (Exception ex)
            {
                //I SHOULD PROBABLY LOG SOMETHING... LIKE AN ERROR =/ BUT HERE'S SOME WORDS INSTEAD... CHILL IT'S A GAME NOT N.A.S.A.
            }
        }

        // calculateAndApplyDrag Method: Calculates drag and aplies it to the grid
        private void calculateAndApplyDrag()
        {
            try
            {
                var gridCenter = grid.Physics.CenterOfMassWorld;

                if (++slowUpdate % SKIP_TICKS == 0)
                {
                    slowUpdate = 0;
                    atmosphere = 0;
                    reEntryAtmosphere = 0;
                    atmospheres = 0;

                    foreach (var kv in AtmospereDetection.planets)
                    {
                        var planet = kv.Value;

                        if (planet.Closed || planet.MarkedForClose)
                        {
                            AtmospereDetection.removePlanets.Add(kv.Key);
                            continue;
                        }

                        if (planet.HasAtmosphere && Vector3D.DistanceSquared(gridCenter, planet.WorldMatrix.Translation) < (planet.AtmosphereRadius * planet.AtmosphereRadius))
                        {
                            float atmo = planet.GetAirDensity(gridCenter);
                            atmosphere += atmo;
                            reEntryAtmosphere += atmo;

                            atmospheres++;
                        }
                    }

                    if (AtmospereDetection.removePlanets.Count > 0)
                    {
                        foreach (var id in AtmospereDetection.removePlanets)
                            AtmospereDetection.planets.Remove(id);

                        AtmospereDetection.removePlanets.Clear();
                    }
                    if (atmospheres > 0)
                    {
                        atmosphere /= atmospheres;
                        atmosphere = MathHelper.Clamp((atmosphere - MIN_ATMOSPHERE) / (MAX_ATMOSPHERE - MIN_ATMOSPHERE), 0f, 1f);
                        atmosphere = (float)Math.Pow(atmosphere, 3);
                    }
                }

                if (atmosphere <= 0)
                    return;

                Vector3 forward = Vector3.Normalize(grid.Physics.LinearVelocity);

                float speedSq = grid.Physics.LinearVelocity.LengthSquared();

                float gridSize = grid.GridSize;

                if (gridSize == 2.5)
                {
                    gridSize = gridSize * 2;
                }

                Vector3D shipForward = grid.WorldMatrix.Forward;
                Vector3D shipLeft = grid.WorldMatrix.Left;
                Vector3D shipUp = grid.WorldMatrix.Up;

                // Atmospheric Drag V1.2 Algorithm
                double dragMultiplierFB = 0;
                double dragMultiplierLR = 0;
                double dragMultiplierUD = 0;

                // Calculate drag from forward/backward side
                double dotForward = Vector3D.Dot(shipForward, forward);
                if (dotForward > 0)
                {
                    dragMultiplierFB += dotForward * countForwardBackward;
                }
                else
                {
                    dragMultiplierFB += dotForward * countForwardBackward;
                }

                // Calculate drag from left/right side
                double dotLeft = Vector3D.Dot(shipLeft, forward);
                if (dotLeft > 0)
                {
                    dragMultiplierLR += dotLeft * countLeftRight;
                }
                else
                {
                    dragMultiplierLR += dotLeft * countLeftRight;
                }

                // Calculate drag from up/down side
                double dotUp = Vector3D.Dot(shipUp, forward);
                if (dotUp > 0)
                {
                    dragMultiplierUD += dotUp * countUpDown;
                }
                else
                {
                    dragMultiplierUD += dotUp * countUpDown;
                }

                // Apply drag force to grid
                grid.Physics.AddForce(MyPhysicsForceType.APPLY_WORLD_FORCE, -shipForward * 2f * speedSq * gridSize * (float)dragMultiplierFB * atmosphere, gridCenter, null);
                grid.Physics.AddForce(MyPhysicsForceType.APPLY_WORLD_FORCE, -shipLeft * 2f * speedSq * gridSize * (float)dragMultiplierLR * atmosphere, gridCenter, null);
                grid.Physics.AddForce(MyPhysicsForceType.APPLY_WORLD_FORCE, -shipUp * 2f * speedSq * gridSize * (float)dragMultiplierUD * atmosphere, gridCenter, null);

                /* Atmospheric Drag v1.1 ALGORITHM
                double dragMultiplier = 0;

                // Calculate drag from forward/backward side
                double dotForward = Vector3D.Dot(shipForward, forward);
                if (dotForward > 0) 
                {
                    dragMultiplier += dotForward * countForwardBackward;
                } 
                else 
                {
                    dragMultiplier += -dotForward * countForwardBackward;
                }
                
                // Calculate drag from left/right side
                double dotLeft = Vector3D.Dot(shipLeft, forward);
                if (dotLeft > 0)
                {
                    dragMultiplier += dotLeft * countLeftRight;
                }
                else
                {
                    dragMultiplier += -dotLeft * countLeftRight;
                }

                // Calculate drag from up/down side
                double dotUp = Vector3D.Dot(shipUp, forward);
                if (dotUp > 0)
                {
                    dragMultiplier += dotUp * countUpDown;
                }
                else
                {
                    dragMultiplier += -dotUp * countUpDown;
                }
                    
                // Apply drag force to grid
                grid.Physics.AddForce(MyPhysicsForceType.APPLY_WORLD_FORCE, -forward * speedSq * gridSize * 1.7f * (float)dragMultiplier * atmosphere, gridCenter, null);
                */
            }
            catch (Exception ex)
            {
                //I SHOULD PROBABLY LOG SOMETHING... LIKE AN ERROR =/ BUT HERE'S SOME WORDS INSTEAD... CHILL IT'S A GAME NOT N.A.S.A.
            }
        }

        // calculateAndApplyCompressionHeat Method: Calculates and applies the compression heat
        private void calculateAndApplyCompressionHeat()
        {
            if (blockCount != grid.BlocksCount)
            {
                updateSurfaceArea();
                blockCount = grid.BlocksCount;
            }

            Vector3 forward = Vector3.Normalize(grid.Physics.LinearVelocity);
            Vector3D shipForward = grid.WorldMatrix.Forward;
            Vector3D shipLeft = grid.WorldMatrix.Left;
            Vector3D shipUp = grid.WorldMatrix.Up;


            List<List<IMySlimBlock>> heatedSurfaces = new List<List<IMySlimBlock>>();
            float heatValue1 = 0;
            float heatValue2 = 0;
            float heatValue3 = 0;

            // Calculate drag from forward/backward side
            double dotForward = Vector3D.Dot(shipForward, forward);
            if (dotForward > 0)
            {
                heatedSurfaces.Add(forwardSurfaceBlocks);
                heatValue1 = (float)dotForward * reEntryAtmosphere;
            }
            else
            {
                heatedSurfaces.Add(backwardSurfaceBlocks);
                heatValue1 = -(float)dotForward * reEntryAtmosphere;
            }

            // Calculate drag from left/right side
            double dotLeft = Vector3D.Dot(shipLeft, forward);
            if (dotLeft > 0)
            {
                heatedSurfaces.Add(leftSurfaceBlocks);
                heatValue2 = (float)dotLeft * reEntryAtmosphere;
            }
            else
            {
                heatedSurfaces.Add(rightSurfaceBlocks);
                heatValue2 = -(float)dotLeft * reEntryAtmosphere;
            }

            // Calculate drag from up/down side
            double dotUp = Vector3D.Dot(shipUp, forward);
            if (dotUp > 0)
            {
                heatedSurfaces.Add(downSurfaceBlocks);
                heatValue3 = (float)dotUp * reEntryAtmosphere;
            }
            else
            {
                heatedSurfaces.Add(upSurfaceBlocks);
                heatValue3 = -(float)dotUp * reEntryAtmosphere;
            }
            heatSurfaces(heatedSurfaces[0], heatValue1, heatedSurfaces[1], heatValue2, heatedSurfaces[2], heatValue3);
            heatedSurfaces.Clear();
        }

        // heatSurface Method: heats the surface area of a side of ship
        private void heatSurfaces(List<IMySlimBlock> blocksSurface1, float heatValue1, List<IMySlimBlock> blocksSurface2, float heatValue2, List<IMySlimBlock> blocksSurface3, float heatValue3)
        {
            CompressionHeatedBlock block = null;
            heatValue1 = (float)Math.Pow((double)heatValue1, 1.5);
            heatValue2 = (float)Math.Pow((double)heatValue2, 1.5);
            heatValue3 = (float)Math.Pow((double)heatValue3, 1.5);

            HashSet<Vector3I> alreadyHeatedSurfaces = new HashSet<Vector3I>();

            // 1,2,3
            if (heatValue1 > heatValue2 && heatValue1 > heatValue3 && heatValue2 > heatValue3)
            {
                alreadyHeatedSurfaces = heatSurface(blocksSurface1, heatValue1, alreadyHeatedSurfaces);
                alreadyHeatedSurfaces = heatSurface(blocksSurface2, heatValue2, alreadyHeatedSurfaces);
                heatSurface(blocksSurface3, heatValue3, alreadyHeatedSurfaces);
            }

            // 1,3,2
            if (heatValue1 > heatValue2 && heatValue1 > heatValue3 && heatValue3 > heatValue2)
            {
                alreadyHeatedSurfaces = heatSurface(blocksSurface1, heatValue1, alreadyHeatedSurfaces);
                alreadyHeatedSurfaces = heatSurface(blocksSurface3, heatValue3, alreadyHeatedSurfaces);
                heatSurface(blocksSurface2, heatValue2, alreadyHeatedSurfaces);
            }

            // 2,1,3
            if (heatValue2 > heatValue1 && heatValue2 > heatValue3 && heatValue1 > heatValue3)
            {
                alreadyHeatedSurfaces = heatSurface(blocksSurface2, heatValue2, alreadyHeatedSurfaces);
                alreadyHeatedSurfaces = heatSurface(blocksSurface1, heatValue1, alreadyHeatedSurfaces);
                heatSurface(blocksSurface3, heatValue3, alreadyHeatedSurfaces);
            }

            // 2,3,1
            if (heatValue2 > heatValue3 && heatValue2 > heatValue1 && heatValue3 > heatValue1)
            {
                alreadyHeatedSurfaces = heatSurface(blocksSurface2, heatValue2, alreadyHeatedSurfaces);
                alreadyHeatedSurfaces = heatSurface(blocksSurface3, heatValue3, alreadyHeatedSurfaces);
                heatSurface(blocksSurface1, heatValue1, alreadyHeatedSurfaces);
            }

            // 3, 2, 1
            if (heatValue3 > heatValue2 && heatValue3 > heatValue1 && heatValue2 > heatValue1)
            {
                alreadyHeatedSurfaces = heatSurface(blocksSurface3, heatValue3, alreadyHeatedSurfaces);
                alreadyHeatedSurfaces = heatSurface(blocksSurface2, heatValue2, alreadyHeatedSurfaces);
                heatSurface(blocksSurface1, heatValue1, alreadyHeatedSurfaces);
            }

            // 3, 1, 2
            if (heatValue3 > heatValue1 && heatValue3 > heatValue2 && heatValue1 > heatValue2)
            {
                alreadyHeatedSurfaces = heatSurface(blocksSurface3, heatValue3, alreadyHeatedSurfaces);
                alreadyHeatedSurfaces = heatSurface(blocksSurface1, heatValue1, alreadyHeatedSurfaces);
                heatSurface(blocksSurface2, heatValue2, alreadyHeatedSurfaces);
            }
        }

        // heatSurface Method: Heats an individual surface list of blocks and returns list of blocks that have already been heated this round
        private HashSet<Vector3I> heatSurface(List<IMySlimBlock> blocksSurface, float heatValue, HashSet<Vector3I> alreadyHeatedBlocks)
        {
			/* COMMENTED OUT TILL I HAVE TIME TO FIX... API CHANGED
            if (heatValue > 0)
            {
                // you only need to spawn it once and it will spawn particles until you stop it
                if (effect == null && MyParticlesManager.TryCreateParticleEffect(101, out effect, false))
                {
                    float scaleMultiplier = grid.Max.AbsMax() + grid.Min.AbsMin();

                    MatrixD locationMatrix = grid.WorldMatrix;
                    locationMatrix.Translation = grid.Physics.CenterOfMassWorld;
                    effect.WorldMatrix = locationMatrix;
                    effect.Velocity = grid.Physics.LinearVelocity;
                    effect.UserScale = scaleMultiplier * 0.75f;
                    effect.LowRes = true;
                }
            }
            else
            {
                effect.Stop();
                effect = null;
            }
			*/

            CompressionHeatedBlock block = null;
            for (int i = 0; i < blocksSurface.Count; i++)
            {
                if (blocksSurface[i] == null || blocksSurface[i].IsDestroyed)
                    continue;
                float heatMultiplier = grid.Physics.LinearVelocity.Length() - (float)COMPRESSION_DAMAGE_THRESHOLD;

                if (heatMultiplier < 200)
                {
                    heatMultiplier = (heatMultiplier * heatMultiplier) / 200;
                }

                Vector3I blockKey = blocksSurface[i].Position;
                if (compressionHeatedBlocks.TryGetValue(blockKey, out block))
                {
                    if (alreadyHeatedBlocks.Contains(blockKey))
                    {
                        block.heatBlock(heatValue * 7f * heatMultiplier);
                    }
                    else
                    {
                        block.heatBlock(heatValue * 45f * heatMultiplier);
                        alreadyHeatedBlocks.Add(blockKey);
                    }
                }
                else
                {
                    block = new CompressionHeatedBlock(blocksSurface[i], grid, heatShieldBlocks);
                    if (alreadyHeatedBlocks.Contains(blockKey))
                    {
                        block.heatBlock(heatValue * 7f * heatMultiplier);
                    }
                    else
                    {
                        block.heatBlock(heatValue * 45f * heatMultiplier);
                        alreadyHeatedBlocks.Add(blockKey);
                    }
                    compressionHeatedBlocks.Add(blockKey, block);
                }
            }
            return alreadyHeatedBlocks;
        }

        /* V1.3 Heat Surface
        // heatSurface Method:
        private HashSet<Vector3I> heatSurface(List<IMySlimBlock> blocksSurface, float heatValue, HashSet<Vector3I> alreadyHeatedBlocks)
        {
            CompressionHeatedBlock block = null;
            for (int i = 0; i < blocksSurface.Count; i++)
            {
                if (blocksSurface[i] == null || blocksSurface[i].IsDestroyed)
                    continue;

                Vector3I blockKey = blocksSurface[i].Position;
                if (compressionHeatedBlocks.TryGetValue(blockKey, out block))
                {
                    //if (vectorExistsInList(alreadyHeatedBlocks, blockKey))
                    if (alreadyHeatedBlocks.Contains(blockKey))
                    {
                        block.heatBlock(heatValue * 14f * (grid.Physics.LinearVelocity.Length() - (float)COMPRESSION_DAMAGE_THRESHOLD));
                    }
                    else
                    {
                        block.heatBlock(heatValue * 80f * (grid.Physics.LinearVelocity.Length() - (float)COMPRESSION_DAMAGE_THRESHOLD));
                        alreadyHeatedBlocks.Add(blockKey);
                    }
                }
                else
                {
                    block = new CompressionHeatedBlock(blocksSurface[i], grid);
                    compressionHeatedBlocks.Add(blockKey, block);
                }
            }
            return alreadyHeatedBlocks;
        }
        */

        // isShipDragEnabled Method: Checks if the ship drag should be enabled based
        private bool isShipValid()
        {
            if (grid == null || grid.Closed || grid.MarkedForClose || grid.Physics == null || grid.Physics.IsStatic || grid.BlocksCount < 3/* || grid.Physics.LinearVelocity.Length() < 0.5f*/)
            {
                return false;
            }
            else
            {
                return true;
            }
        }

        private void addHeatShieldBlocks()
        {
            //Large Ship
            heatShieldBlocks.Add(16500);
            heatShieldBlocks.Add(8250);
            heatShieldBlocks.Add(2800);
            heatShieldBlocks.Add(14000);
            heatShieldBlocks.Add(14500);
            heatShieldBlocks.Add(13700);
            heatShieldBlocks.Add(8250);
            heatShieldBlocks.Add(3240);
            heatShieldBlocks.Add(14500);
            heatShieldBlocks.Add(13700);
            heatShieldBlocks.Add(15500);

            //Small Ship
            heatShieldBlocks.Add(560);
            heatShieldBlocks.Add(330);
            heatShieldBlocks.Add(230);
            heatShieldBlocks.Add(430);
            heatShieldBlocks.Add(530);
        }
    }
}
